//   This file is part of Scilab Aurora 
//	( http://scilab-aurora.sourceforge.net/ )
//   Copyright (C) 2006  XMU Scilab Aurora Team
//   Authors: (Scilab Aurora Team)
//     Jiang Fengbin <jonbo@126.com>
//     Lin Chuanwen <fjjllcw695@163.com>
//     Xiao Wenqing <pblz.xm@gmail.com>
//     Yang Kaigui  <justforward.xm@gmil.com>
//     Zheng Jiemin 
//     Zhou Xiyue   <zxy_0725@163.com>
//     Lin Shuyang  <macroxmu@gmail.com>
//   This program is licensed under GPL License;
//


// linux  version using xlib & gdk-pixbuf
// imdraw(img, id, x, y ,w, h )
// draw image to window with id
// for some reason, linux version ignore x,y,w,h field,

#include <stdio.h>
#include <string.h>

#include "stack-c.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>

#include <gdk-pixbuf/gdk-pixbuf.h>

#include <unistd.h>

int already_init = 0;
Display * display;
int screen_num;
GC gc;
// Window window;
unsigned char  *buf;	// pre alloced buf

// int _x, _y, _w, _h;	// preview area

char msg_buf[256];
inline void debug(char *msg)
{
//	sciprint (msg);
//	sciprint ("\n");
//	FILE *fp = fopen("debug.txt", "a");
//	fprintf( fp, "%s\n", msg );
//	fclose( fp );
}

// system depend draw operation
// linux verion: x,y,w,h is ignored
int imdraw(int win, int width,int height,char *pData,int x,int y,int w,int h)
{
	int width_step;
	char *pDst, *pScr;
	int row, col, ch;
	int i;
	
if (already_init==0) {
	already_init = 1;
	
	//  connect to x-server , init libs, alloc buffer
	if ( (display=XOpenDisplay(NULL)) == NULL) {
		error("Can't connect to X server");
		return -1;
	}
	screen_num = DefaultScreen(display);
	gc = DefaultGC(display,screen_num);
	gdk_pixbuf_xlib_init( display,screen_num );
	xlib_rgb_init( display,screen_num );
}
	
	buf = (unsigned char*) malloc (width * height * 3);
	if (!buf) { 
		debug (" error when alloc"); 
		return -1;
	}
	
	width_step = width*3;
	pDst = buf;	pScr = pData;
	for(ch = 0; ch < 3 ; ch++)
		for(col =0; col < width; col++)
			for(row = 0; row < height; row++){
				// (x,y)=(col,row)
				*(pDst + width_step*row + col*3 + ch) = *pScr++;
			}
	xlib_draw_rgb_image ( win, gc, 0, 0, width, height, 0, buf, width_step );
	XFlush(display);
	
	free(buf);
	return 0;
}

// check types & get width, height ,Data
int get_data( int pos, int *pWidth, int *pHeight, char **ppData)
{
  char ** pStr;
  int * pListHeader;
  int * pDataHeader;
  int m1, n1, m2, n2, m3, n3;
  SciIntMat Dims;
  SciIntMat IntMat;
  int nWidth, nHeight, nCh;

  //check whether the the argument is a hypermatrix
	if( VarType(pos) != 17) // not a hypermatrix
  	return -1;
	GetListRhsVar (pos, 1, "S", &m1, &n1, &pStr);
  if( m1 !=1 || n1 !=3)
  	return -1;
  if( strcmp(pStr[0], "hm") != 0 || strcmp(pStr[1], "dims") != 0 || strcmp(pStr[2], "entries") != 0)
  	return -1;

  //get the dimension information, it's stored in the second element of the mlist
  GetListRhsVar(pos, 2, "I", &m2, &n2, &Dims);
  if( m2*n2 !=3)
  	return -1;
  nHeight = IC_INT32(Dims.D)[0];
  nWidth  = IC_INT32(Dims.D)[1];
  nCh = IC_INT32(Dims.D)[2];
  if (nCh!=3)
  	return -1;
  
  //get mlist data structure
  pListHeader = (int*)GetData(pos);
  //data is stored in the third element
 //data struct is get first
  pDataHeader = (int*)( ((char*)pListHeader) + 24 + (pListHeader[4]-1)*8 );
  //the data type is store in the first 4 bytes as a integer
  if ( pDataHeader[0]!=8 )	// data type must be integer
  	return -1;
  
	GetListRhsVar(pos, 3, "I", &m3, &n3, &IntMat);
	m3 = IntMat.m; 	n3 = IntMat.n;
	if (IntMat.it!=I_UCHAR)
  	return -1;
	if(m3*n3 != nWidth * nHeight * nCh) 	//check dimension 
  	return -1;
	// pData = IntMat.D;

	// now we got nWidth, nHeight, pData
	*pWidth = nWidth;
	*pHeight = nHeight;
	*ppData = (char*)IntMat.D;
 
EXIT:
//  FreeRhsSVar(pStr);
  // <==bug here?, do I have to free other vars?
  return 0;
}


/////////////////////////////////////////////////
// steps:
// checking inputs must be a M*N*3 hypermatrix (type is uint8)
// get image dimension information and the data address
// get correct hwd & hdc for draw, create a new thread and window if neccesary
// convert Scilab 3D Mat to Dib
// calling StretchDIBits to draw
int int_imdraw( char *fname)
{
	// int m1, n1, l1;
	int m2, n2, l2;
	int m3, n3, l3;
	int m4, n4, l4;
	int m5, n5, l5;
	int m6, n6, l6;
 
	// checking inputs
	CheckRhs(6, 6) ;
	CheckLhs(0, 1) ;
	// GetRhsVar(1,"d",&m1,&n1,&l1);
	// GetRhsVar(2,"d",&m2,&n2,&l2);
	GetRhsVar(2,"c",&m2,&n2,&l2);
	GetRhsVar(3,"d",&m3,&n3,&l3);
	GetRhsVar(4,"d",&m4,&n4,&l4);
	GetRhsVar(5,"d",&m5,&n5,&l5);
	GetRhsVar(6,"d",&m6,&n6,&l6);
	
	// get position args
	int id, x, y ,w ,h;
	sscanf( cstk(l2), "%i", &id );
	// id = (int)*stk(l2);
	x = (int) *stk(l3);
	y = (int) *stk(l4);
	w = (int) *stk(l5);
	h = (int) *stk(l6);
		
	// check and get data
	int width, height;
	char *pData;

	if ( get_data(1, &width, &height, &pData)<0 ){
		sciprint("Not a RGB image?");
		return 0;	
	}
	
	imdraw(id,width,height,pData,x,y,w,h);

	return 0;
}

// init preview window
// init_preview(int x, int y, int w, int h)
int int_init_preview( char *fname)
{
/*
	int m1, n1, l1;
	int m2, n2, l2;
	int m3, n3, l3;
	int m4, n4, l4;
 
	// checking inputs
	CheckRhs(4, 4) ;
	CheckLhs(0, 1) ;
	GetRhsVar(1,"d",&m1,&n1,&l1);
	GetRhsVar(2,"d",&m2,&n2,&l2);
	GetRhsVar(3,"d",&m3,&n3,&l3);
	GetRhsVar(4,"d",&m4,&n4,&l4);
	
	// get preview area
	_x = (int)*stk(l1);
	_y = (int)*stk(l2);
	_w = (int)*stk(l3);
	_h = (int)*stk(l4);

	//  connect to x-server , init libs, alloc buffer
	if ( (display=XOpenDisplay(NULL)) == NULL) {
		error("Can't connect to X server");
		return -1;
	}
	screen_num = DefaultScreen(display);
	gc = DefaultGC(display,screen_num);
	gdk_pixbuf_xlib_init( display,screen_num );
	xlib_rgb_init( display,screen_num );
	buf = (unsigned char*) malloc (_w * _h * 3);
	if (!buf) { 
		debug (" error when alloc"); 
		return -1;
	}
	
	// create window	
	window = XCreateSimpleWindow(display, RootWindow(display,screen_num) , 
		_x, _y, _w, _h,	0, BlackPixel(display,screen_num),WhitePixel(display,screen_num) );
		
	unsigned long valuemask = 0;
	XGCValues values;
	gc = XCreateGC(display, window, valuemask, &values);
	
	XSetWindowAttributes attr;
	unsigned long amask; 
	attr.override_redirect=True;
	amask=CWOverrideRedirect;
	XChangeWindowAttributes(display,window,amask,&attr);
	
	XMapWindow(display, window);
	XFlush(display);
	// XDrawRectangle(display, window, gc, 10, 10, 100, 100);
*/
	return 0;
}


int int_close_preview( char *fname)
{
/*
	if (!buf) return 0;
	free(buf);
	XFreeGC(display, gc);
	XCloseDisplay(display);
	return 0;
*/
}

