//   This file is part of Scilab Aurora 
//	( http://scilab-aurora.sourceforge.net/ )
//   Copyright (C) 2006  XMU Scilab Aurora Team
//   Authors: (Scilab Aurora Team)
//     Jiang Fengbin <jonbo@126.com>
//     Lin Chuanwen <fjjllcw695@163.com>
//     Xiao Wenqing <pblz.xm@gmail.com>
//     Yang Kaigui  <justforward.xm@gmil.com>
//     Zheng Jiemin 
//     Zhou Xiyue   <zxy_0725@163.com>
//     Lin Shuyang  <macroxmu@gmail.com>
//   This program is licensed under GPL License;
//


// windows version
// imdraw(img, [mod, x, y ,w, h] )
// mod: default is 0, pop up a window, and show
//      when mod = 1, draw the img direct to screen, position at (x,y,w,h)
//      when mod = 2, pop up a window with postion (x,y,w,h),
//                    and show img in the win

extern "C"{
#include <windows.h>	
#include "stack-c.h"
#include <stdio.h>
#include <string.h>

inline void debug(char *msg)
{
	return;
//	FILE *fp = fopen("debug.txt", "a");
//	fprintf( fp, "%s\n", msg );
//	fclose( fp );
}

// check types & get width, height ,Data
int get_data( int pos, int *pWidth, int *pHeight, char **ppData)
{
  char ** pStr;
  int * pListHeader;
  int * pDataHeader;
  int m1, n1, m2, n2, m3, n3;
  SciIntMat Dims;
  SciIntMat IntMat;
  int nWidth, nHeight, nCh;

	debug(" checking is it a hypermatrix" );
  //check whether the the argument is a hypermatrix
	if( VarType(pos) != 17) // not a hypermatrix
  	return -1;
	GetListRhsVar (pos, 1, "S", &m1, &n1, &pStr);
  if( m1 !=1 || n1 !=3)
  	return -1;
  if( strcmp(pStr[0], "hm") != 0 || strcmp(pStr[1], "dims") != 0 || strcmp(pStr[2], "entries") != 0)
  	return -1;

	debug("get the dimension information");
  //get the dimension information, it's stored in the second element of the mlist
  GetListRhsVar(pos, 2, "I", &m2, &n2, &Dims);
  if( m2*n2 !=3)
  	return -1;
  nHeight = IC_INT32(Dims.D)[0];
  nWidth  = IC_INT32(Dims.D)[1];
  nCh = IC_INT32(Dims.D)[2];
  if (nCh!=3)
  	return -1;
  
  debug("get mlist data structure");	
  //get mlist data structure
  pListHeader = (int*)GetData(pos);
  //data is stored in the third element
 //data struct is get first
  pDataHeader = (int*)( ((char*)pListHeader) + 24 + (pListHeader[4]-1)*8 );
  //the data type is store in the first 4 bytes as a integer
  if ( pDataHeader[0]!=8 )	// data type must be integer
  	return -1;
  
  debug(" get dimesion");	
	GetListRhsVar(pos, 3, "I", &m3, &n3, &IntMat);
	m3 = IntMat.m; 	n3 = IntMat.n;
	if (IntMat.it!=I_UCHAR)
  	return -1;
	if(m3*n3 != nWidth * nHeight * nCh) 	//check dimension 
  	return -1;
	// pData = IntMat.D;

	// now we got nWidth, nHeight, pData
	*pWidth = nWidth;
	*pHeight = nHeight;
	*ppData = (char*)IntMat.D;
 
EXIT:
  debug(" return from get_data" );
//  FreeRhsSVar(pStr);
  // <==bug here?, do I have to free other vars?
  return 0;
}


// show img at hdc's (x,y,w,h)
int imdraw( HDC hdc, int width, int height, char *pData, int x, int y, int w, int h )
{
	BITMAPINFO  bitmapInfo;
	memset(&bitmapInfo,0,sizeof(bitmapInfo));
	bitmapInfo.bmiHeader.biSize = 40;
	bitmapInfo.bmiHeader.biPlanes = 1;
	bitmapInfo.bmiHeader.biBitCount = 24;
	bitmapInfo.bmiHeader.biCompression = 0;// BI_RGB;
	bitmapInfo.bmiHeader.biXPelsPerMeter = 0;
	bitmapInfo.bmiHeader.biYPelsPerMeter = 0;
	bitmapInfo.bmiHeader.biClrUsed = 0;
	bitmapInfo.bmiHeader.biClrImportant = 0;
	
	bitmapInfo.bmiHeader.biWidth = width;
	bitmapInfo.bmiHeader.biHeight = height;
	int dwWidthBytes = width*3;
  while((dwWidthBytes&3)) 
  	dwWidthBytes++;
	bitmapInfo.bmiHeader.biSizeImage = dwWidthBytes*height;

//	bitmapInfo.bmiHeader.biSizeImage = 3*width*height;
	
	debug("alloc tmp buffer");
	char *tmp;
	tmp = (char*) malloc( dwWidthBytes*height );
	if (!tmp) // alloc error
		return -1;

	debug(" copying data");
	// now  it's time to copy data
	char *pDst = tmp;
	char *pScr = pData;
	int row, col, ch, widthStep;
	for(ch = 0; ch < 3 ; ch++)
		for(col =0; col < width; col++)
			for(row = 0; row < height; row++){
				// (x,y)=(col,row)
				*(pDst + dwWidthBytes*row + col*3 + 2-ch) = *pScr++;
			}
	
	debug("draw it");
	// draw it
	StretchDIBits(hdc, 
			x, y+h, w, -h,
			0, 0, width, height, 
			tmp, &bitmapInfo, DIB_RGB_COLORS,SRCCOPY);
	
	debug("return from draw");
	// clear up
	free( tmp );
	return 0;
}


// steps:
// checking inputs must be a M*N*3 hypermatrix (type is uint8)
// get image dimension information and the data address
// get correct hwd & hdc for draw, create a new thread and window if neccesary
// convert Scilab 3D Mat to Dib
// calling StretchDIBits to draw
int int_imdraw( char *fname)
{
	// int m1, n1, l1;
	int m2, n2, l2;
	int m3, n3, l3;
	int m4, n4, l4;
	int m5, n5, l5;
	int m6, n6, l6;
 
	debug("checking inputs");
	// checking inputs
	CheckRhs(6, 6) ;
	CheckLhs(0, 1) ;
	// GetRhsVar(1,"d",&m1,&n1,&l1);
	GetRhsVar(2,"d",&m2,&n2,&l2);
	GetRhsVar(3,"d",&m3,&n3,&l3);
	GetRhsVar(4,"d",&m4,&n4,&l4);
	GetRhsVar(5,"d",&m5,&n5,&l5);
	GetRhsVar(6,"d",&m6,&n6,&l6);
	
	debug("get position args");
	// get position args
	int mod, x, y ,w ,h;
	mod = (int)*stk(l2);
	x = (int) *stk(l3);
	y = (int) *stk(l4);
	w = (int) *stk(l5);
	h = (int) *stk(l6);
	
	debug("check and get data");
	// check and get data
	int width, height;
	char *pData;

	if ( get_data(1, &width, &height, &pData)<0 ){
		sciprint("Not a RGB image?");
		return 0;	
	}
	
	debug("convert to Did");
	// convert to Did
	imdraw( GetDC(0), width, height, pData, x, y, w, h );

	debug("return from int_indraw");
	return 0;
}

};// end of extern "C"
