//   This file is part of Scilab Aurora 
//	( http://scilab-aurora.sourceforge.net/ )
//   Copyright (C) 2006  XMU Scilab Aurora Team
//   Authors: (Scilab Aurora Team)
//     Jiang Fengbin <jonbo@126.com>
//     Lin Chuanwen <fjjllcw695@163.com>
//     Xiao Wenqing <pblz.xm@gmail.com>
//     Yang Kaigui  <justforward.xm@gmil.com>
//     Zheng Jiemin 
//     Zhou Xiyue   <zxy_0725@163.com>
//     Lin Shuyang  <macroxmu@gmail.com>
//   This program is licensed under GPL License;
//



// image I/O for scilab using opencv
// we only support RGB image
// by Jonbo
// im = imread( filename )
// imwrite(im, filename)

#ifdef __cplusplus
extern "C" {
#endif

#define CV_INLINE static
#ifndef TRUE
    #define TRUE 1
#endif
#ifndef FALSE
    #define FALSE 0
#endif
typedef int BOOL;
#define SIVP_FLOAT 32
#define SIVP_DOUBLE 64

#include <stdio.h>
#include <cv.h>
#include <cxcore.h>
#include <highgui.h>
#include <stack-c.h>
#include <sys/types.h>
#include <sys/stat.h>
//#include <unistd.h>

void debug(char *msg)
{
	
}

// save image to scilab hypermatrix
int RGBImg2Sci(IplImage * pImage, int nPos)
{
  static char *Str[]= { "hm","dims","entries"}; 
  int m1=1,n1=3;
  int mL=3,nL=1,lL, un=1;
  const int nCh = 3;
  int nRow = pImage->height ;
  int nCol = pImage->width ;
  int d[3];
  SciIntMat Dims;
  SciIntMat IntData;
  char *pSrc,*pDst; 
  int ch, row, col;
 	int widthStep;

  //create SciIntMat for dimentional information
  Dims.m = 1;
  Dims.n = 3;
  Dims.l = -1; /* dimensions matrix not in scilab stack */
  Dims.it = I_INT32;
	Dims.D = d;
	d[0] = nRow;
	d[1] = nCol;
	d[2] = nCh;

  //create SciIntMat for the image data
  IntData.m = nRow ;
  IntData.n = nCol * nCh;
  IntData.l = -1; /* dimensions matrix not in scilab stack */
  IntData.it = I_UCHAR;//nType;
	IntData.D = malloc( pImage->width * pImage->height * 3 * 1 );
	if (!IntData.D){
		return -1;
	}
	
	// copy data	
  pSrc = (char*)(char*)(pImage->imageData);
  pDst = (char*)IntData.D;
	widthStep = pImage->widthStep;

  for(ch = 0; ch < nCh ; ch++){ //the order of IplImage is BGR
    for(col =0; col < nCol; col++){
      for(row = 0; row < nRow; row++){
				// (x,y) = (col, row)
				*pDst++ = * (pSrc + row*widthStep + col*nCh + 2-ch);
			}
		}
	}
	
  CreateVar(nPos,"m", &mL, &nL, &lL);
  CreateListVarFromPtr(nPos,1,"S", &m1, &n1, Str);
  CreateListVarFromPtr(nPos,2,"I",&(Dims.m), &(Dims.n), &Dims); 
  CreateListVarFromPtr(nPos,3,"I",&(IntData.m), &(IntData.n), &IntData ); 

// free(Dims.D);
	free( IntData.D );
  return 0;
}

// im = imread( filename )
int int_imread(char * fname)
{
	int mR, nR, lR;
	IplImage * pImage;
	int nBytes;
	
	CheckRhs(1, 1);
	CheckLhs(1, 1);
	GetRhsVar(1, "c", &mR, &nR, &lR);

	pImage = cvLoadImage(cstk(lR), -1);
	if(pImage == NULL){/* if load image failed */
		sciprint("Can not open file %s.\r\n", cstk(lR));
		return -1;
  }
  
  // check, we only support RGB image, to unique the internal data struct for image
  /*how many bytes per pixel per channel*/
  nBytes = pImage->depth;
  if (nBytes > IPL_DEPTH_SIGN)
    nBytes -= IPL_DEPTH_SIGN;
  nBytes = nBytes >> 3;

  if ( pImage->nChannels!=3 || nBytes!=1 ){
  	sciprint("we only support RGB image!\n");
  	return -1;
  }
  
  // save img to siclab hypermatrix
	RGBImg2Sci(pImage, 2);
	// clear up & return
	cvReleaseImage(&pImage);
	LhsVar(1) = 2;
	return 0;
}

//****************************** imwrite ******************************

// check types & get width, height ,Data
int get_data( int pos, int *pWidth, int *pHeight, char **ppData)
{
  char ** pStr;
  int * pListHeader;
  int * pDataHeader;
  int m1, n1, m2, n2, m3, n3;
  SciIntMat Dims;
  SciIntMat IntMat;
  int nWidth, nHeight, nCh;

	debug(" checking is it a hypermatrix" );
  //check whether the the argument is a hypermatrix
	if( VarType(pos) != 17) // not a hypermatrix
  	return -1;
	GetListRhsVar (pos, 1, "S", &m1, &n1, &pStr);
  if( m1 !=1 || n1 !=3)
  	return -1;
  if( strcmp(pStr[0], "hm") != 0 || strcmp(pStr[1], "dims") != 0 || strcmp(pStr[2], "entries") != 0)
  	return -1;

	debug("get the dimension information");
  //get the dimension information, it's stored in the second element of the mlist
  GetListRhsVar(pos, 2, "I", &m2, &n2, &Dims);
  if( m2*n2 !=3)
  	return -1;
  nHeight = IC_INT32(Dims.D)[0];
  nWidth  = IC_INT32(Dims.D)[1];
  nCh = IC_INT32(Dims.D)[2];
  if (nCh!=3)
  	return -1;
  
  debug("get mlist data structure");	
  //get mlist data structure
  pListHeader = (int*)GetData(pos);
  //data is stored in the third element
 //data struct is get first
  pDataHeader = (int*)( ((char*)pListHeader) + 24 + (pListHeader[4]-1)*8 );
  //the data type is store in the first 4 bytes as a integer
  if ( pDataHeader[0]!=8 )	// data type must be integer
  	return -1;
  
  debug(" get dimesion");	
	GetListRhsVar(pos, 3, "I", &m3, &n3, &IntMat);
	m3 = IntMat.m; 	n3 = IntMat.n;
	if (IntMat.it!=I_UCHAR)
  	return -1;
	if(m3*n3 != nWidth * nHeight * nCh) 	//check dimension 
  	return -1;
	// pData = IntMat.D;

	// now we got nWidth, nHeight, pData
	*pWidth = nWidth;
	*pHeight = nHeight;
	*ppData = (char*)IntMat.D;
 
  debug(" return from get_data" );
//  FreeRhsSVar(pStr);
  // <==bug here?, do I have to free other vars?
  return 0;
}

IplImage * Sci2RGBImg( int pos )
{
	int width, height;
	char *pData;
  IplImage * pImage = NULL;
	char *pDst, *pScr;
	int row, col, ch, widthStep;

	// check and get data
	if ( get_data( pos, &width, &height, &pData)<0 ){
		sciprint("Not a RGB image?");
		return NULL;	
	}
	
  /*create a IplImage to receiving the data*/
  pImage = cvCreateImage(cvSize(width, height), IPL_DEPTH_8U , 3);
  if(pImage == NULL) {
    sciprint("Create IplImage for %d'th argument failed.\r\n", pos);
    return NULL;
  }
    

	// now  it's time to copy data
	pDst = (char*)(pImage->imageData);
	pScr = (char*) pData;
	widthStep = pImage->widthStep;
	for(ch = 0; ch < 3 ; ch++)
		for(col =0; col < width; col++)
			for(row = 0; row < height; row++){
				// (x,y)=(col,row)
				*(pDst + widthStep*row + col*3 + 2-ch) = *pScr++;
			}
			
	return pImage;
	
}

// ret = imwrite( im, filename )
int int_imwrite( char *fname )
{
	int m1, n1, l1;
	int mL, nL, lL;
	char *filename;
  IplImage * pImage;
	
	CheckRhs(2, 2) ;
	CheckLhs(0, 1) ;

  // get file name
  GetRhsVar(2, "c", &m1, &n1, &l1);
  filename = cstk(l1);

  //create return value
  // CreateVar(3, "d", &mL, &nL, &lL);
  // *stk(lL) = -1;

	// convert the matrix to IplImage
  pImage = Sci2RGBImg(1);
  if (!pImage){
  	sciprint("error when create IplImage");
  	return 0;
  }

	// save image
	*stk(lL) = cvSaveImage( filename, pImage);

	// clear up & return
  cvReleaseImage(&pImage);
  
  // LhsVar(1) = 3;
	return 0;
}

#ifdef __cplusplus
}
#endif
