//   This file is part of Scilab Aurora 
//	( http://scilab-aurora.sourceforge.net/ )
//   Copyright (C) 2006  XMU Scilab Aurora Team
//   Authors: (Scilab Aurora Team)
//     Jiang Fengbin <jonbo@126.com>
//     Lin Chuanwen <fjjllcw695@163.com>
//     Xiao Wenqing <pblz.xm@gmail.com>
//     Yang Kaigui  <justforward.xm@gmil.com>
//     Zheng Jiemin 
//     Zhou Xiyue   <zxy_0725@163.com>
//     Lin Shuyang  <macroxmu@gmail.com>
//   This program is licensed under GPL License;
//


// video open for read using ffmpeg by Jonbo
// n = video_open( filename)
// [rate,num] = video_info( n )
// img = video_read_frame( n )
// video_close(n)
// video_close_all()
// ret = video_seek( frame )

extern "C"{
	
#include <avcodec.h>
#include <avformat.h>
#include <stdio.h>
#include "stack-c.h"
#include <string.h>


char buf[256];
inline void debug(char *msg)
{
	// fprintf( stdout, "%s\n", msg);
	// fflush(stdout);
	// sciprint( msg );
	// sciprint( "\n" );
}


int already_registered = 0;		// register only once

// opened video file table
const int MAX_VIDEO_FILE_NUM = 20;

typedef struct OpenedVideo{
	AVFormatContext *pFormatCtx;	// media file handle?
	AVCodecContext  *pCodecCtx;	//
	AVCodec *pCodec;	// opened codec
	uint8_t *buffer;	// buffer for pFrameRGB
	AVFrame *pFrameRGB;	// alloc struct of AVFrame for img convert
	AVFrame *pFrame; 	// alloc struct of AVFrame for get Frame
	int videoStream;	//instested stream id
	uint8_t *sci_buf;	// buffer for converting to Scilab's Mat
	// addition: video informations
	long cur_frame;
	long frame_num;
	double frame_rate;
} OpenedVideo;

OpenedVideo OpenedV[MAX_VIDEO_FILE_NUM];


// skip a video frame, return -1, if got to the end of file
int skip_frame(	
	AVFormatContext *pFormatCtx,
	AVCodecContext  *pCodecCtx,
	AVFrame         *pFrame,
	int             videoStream)
{
	int ret = -1;
	AVPacket packet;
	int frameFinished = 0;

	// Read frames and save first five frames to disk
	frameFinished = 0;
	while ( frameFinished==0 ) {
		if (av_read_frame(pFormatCtx, &packet)>=0) { //get a frame
			// Is this a packet from the video stream?
			if(packet.stream_index==videoStream){
				// Decode video frame
				avcodec_decode_video(pCodecCtx, pFrame, &frameFinished, packet.data, packet.size);
				// Did we get a video frame?
				if(frameFinished){
					ret = 1;
				}
			}
			// Free the packet that was allocated by av_read_frame
			av_free_packet(&packet);
	
		}else { 
			// end of file, return the last frame, bug? if no frame is readed
			frameFinished = 1;
			ret = -1;
		}
	} // end of while
	return ret;
}


/*
get_num()
{
	open video
	skip all frame, frame_num++;
	close video
	return num;
}
*/

// get video file's frame_num, or -1 if error
long get_num( char *filename )
{
	AVFormatContext *pFormatCtx;
	AVCodecContext  *pCodecCtx;
	AVCodec         *pCodec;
	AVFrame         *pFrame; 
	int             videoStream;
	
	// Open video file
	if(av_open_input_file(&pFormatCtx, filename, NULL, 0,  NULL)!=0)
		return -1; // Couldn't open file

	// Retrieve stream information
	if ( av_find_stream_info(pFormatCtx)<0 )
	    return -1; // Couldn't find stream information
	
	// Find the first video stream
	videoStream=-1;
	for(int i=0; i<pFormatCtx->nb_streams; i++)
	    if(pFormatCtx->streams[i]->codec->codec_type==CODEC_TYPE_VIDEO){
	        videoStream=i;
	        break;
	    }
	if(videoStream==-1)
		return -1; // Didn't find a video stream

	// Get a pointer to the codec context for the video stream
	pCodecCtx= pFormatCtx->streams[videoStream]->codec;
	
	// Find the decoder for the video stream
	pCodec=avcodec_find_decoder(pCodecCtx->codec_id);
	if(pCodec==NULL)
	    return -1; // Codec not found
	
	// Open codec
	if(avcodec_open(pCodecCtx, pCodec)<0)
	    return -1; // Could not open codec
	
	// Hack to correct wrong frame rates that seem to be generated by some 
	// codecs
	//    if(pCodecCtx->frame_rate>1000 && pCodecCtx->frame_rate_base==1)
	//        pCodecCtx->frame_rate_base=1000;
	
	// Allocate video frame
	pFrame = avcodec_alloc_frame();

	// skip all the video frames, get frame_num
	int frame_num = 0;
	while( skip_frame( pFormatCtx, pCodecCtx, pFrame, videoStream )>= 0 ){
		frame_num ++;
	}

	// Free the YUV frame
	av_free(pFrame);
	// Close the codec
	avcodec_close(pCodecCtx);
	// Close the video file
	av_close_input_file(pFormatCtx);
	return frame_num;
}


// n = video_open( filename )
int video_open( int slot, char *filename);
// parsing input, find a slot, call video_open
int int_video_open(char *fname)
{
	int l1, m1, n1;
	int l2, m2, n2;
	int n, ret;
	int *pret = &n;
	m2=n2=1;
	
	CheckRhs(1,1) ;
	CheckLhs(1,1) ;
	GetRhsVar(1, "c", &m1, &n1, &l1); 

	// find a empty slot
	for ( n = 0; n < MAX_VIDEO_FILE_NUM; n++){
		if( !(OpenedV[n].pFormatCtx))
			break;
  }
	if( n ==  MAX_VIDEO_FILE_NUM){
		sciprint("%s: Too many video files  opened.  close some files .\r\n", fname);
		// return -1;
		n = -1;	goto _ERR;
	}
	
	OpenedV[n].pFormatCtx = 0;
	
	// open video & save enviroment
	ret = video_open( n, cstk(l1) );
	if ( ret<0 ){
		sciprint(  "%s: Can not open video file.\r\n", fname);
		n = -1; 
	}
	
_ERR:
	// return handle
	m2=n2=1;
	CreateVarFromPtr(2, "i", &m2, &n2, &pret);
	LhsVar(1) = 2;
	return 0;
}

// open file, get information, find streams
// open codec, alloc buf/frame
// fill the slot with video informations
int video_open( int slot, char *filename )
{
	// vars that will be saved
	AVFormatContext *pFormatCtx;
	AVCodec         *pCodec;
	AVFrame         *pFrame; 
	AVFrame         *pFrameRGB;
	uint8_t         *buffer;
	int             videoStream;
	uint8_t *sci_buf;

	//loacl vars
	int i;
	AVCodecContext  *pCodecCtx;
	int             numBytes;
	AVFormatParameters params;
		
	// Register all formats and codecs
	if (already_registered == 0){
	  av_register_all();
	  already_registered = 1;
	}
	
	// get frmae number
	long frame_num = get_num( filename );
	if (frame_num<0) 
		return -1;
	
	// Open video file
	if(av_open_input_file(&pFormatCtx, filename, NULL, 0,  NULL)!=0)
		return -1; // Couldn't open file

	// Retrieve stream information
	if ( av_find_stream_info(pFormatCtx)<0 )
	    return -1; // Couldn't find stream information
	
	// Dump information about file onto standard error
	// dump_format(pFormatCtx, 0, filename, false);
	
	// Find the first video stream
	videoStream=-1;
	for(i=0; i<pFormatCtx->nb_streams; i++)
	    if(pFormatCtx->streams[i]->codec->codec_type==CODEC_TYPE_VIDEO){
	        videoStream=i;
	        break;
	    }
	if(videoStream==-1)
		return -1; // Didn't find a video stream

	// Get a pointer to the codec context for the video stream
	pCodecCtx= pFormatCtx->streams[videoStream]->codec;
	
	// Find the decoder for the video stream
	pCodec=avcodec_find_decoder(pCodecCtx->codec_id);
	if(pCodec==NULL)
	    return -1; // Codec not found
	
	// Open codec
	if(avcodec_open(pCodecCtx, pCodec)<0)
	    return -1; // Could not open codec
	
	// Hack to correct wrong frame rates that seem to be generated by some 
	// codecs
	//    if(pCodecCtx->frame_rate>1000 && pCodecCtx->frame_rate_base==1)
	//        pCodecCtx->frame_rate_base=1000;
	
	// Allocate video frame
	pFrame = avcodec_alloc_frame();
	
	// Allocate an AVFrame structure
	pFrameRGB=avcodec_alloc_frame();
	if(pFrameRGB==NULL)
	    return -1;
	
	// Determine required buffer size and allocate buffer
	numBytes=avpicture_get_size(PIX_FMT_RGB24, pCodecCtx->width, pCodecCtx->height);
	buffer = (uint8_t*)malloc(numBytes);
	if (!buffer) 	return -1;
	sci_buf = (uint8_t*)malloc(numBytes);
	if (!sci_buf)	return -1;
	
	// Assign appropriate parts of buffer to image planes in pFrameRGB
	avpicture_fill((AVPicture *)pFrameRGB, buffer, PIX_FMT_RGB24, pCodecCtx->width, pCodecCtx->height);

	// compute rate
	AVStream *st = 	pFormatCtx->streams[videoStream];
	double frame_rate;
	if(st->r_frame_rate.den && st->r_frame_rate.num)
		frame_rate =  av_q2d(st->r_frame_rate);
	else
		frame_rate =  1/av_q2d(st->codec->time_base);
	
	// save enviroment
	OpenedV[slot].pFormatCtx = pFormatCtx ;
	OpenedV[slot].pCodec = pCodec;
	OpenedV[slot].buffer = buffer;
	OpenedV[slot].pFrameRGB = pFrameRGB;
	OpenedV[slot].pFrame = pFrame; 
	OpenedV[slot].videoStream = videoStream;
	OpenedV[slot].pCodecCtx = pCodecCtx;
	OpenedV[slot].sci_buf = sci_buf;
	// init some video informations
	OpenedV[slot].frame_num = frame_num;
	OpenedV[slot].cur_frame = 0;
	OpenedV[slot].frame_rate = frame_rate;
	
	return 0;
}


// video_close(n);
int video_close( int slot );
int int_video_close(char *fname)
{
	int l1, m1, n1;
	CheckRhs(1,1) ;
	CheckLhs(0,1) ;
	GetRhsVar(1, "d", &m1, &n1, &l1); 
	
	if ( video_close( (int)*stk(l1)) <0 ){
		sciprint("error when closing, maybe the file is not open?");
		// return -1;
	}
	return 0;
}

int video_close( int slot )
{
	// set up enviroment from table
	AVFormatContext *pFormatCtx = OpenedV[slot].pFormatCtx ;
	AVCodec	*pCodec = OpenedV[slot].pCodec;
	uint8_t *buffer = OpenedV[slot].buffer;
	AVFrame *pFrameRGB = OpenedV[slot].pFrameRGB;
	AVFrame *pFrame = OpenedV[slot].pFrame; 
	AVCodecContext  *pCodecCtx =  OpenedV[slot].pCodecCtx;
	uint8_t* sci_buf =  OpenedV[slot].sci_buf;
	// do some checking 
	if (pFormatCtx==0) return -1;	// not open or close, return 
	
	// free alloced buffer
	free( buffer );
	free( sci_buf );
	
	// Free the RGB image
	av_free(pFrameRGB);

	// Free the YUV frame
	av_free(pFrame);

	// Close the codec
	avcodec_close(pCodecCtx);

	// Close the video file
	av_close_input_file(pFormatCtx);

	// clear up slot
	OpenedV[slot].pFormatCtx = 0;
	return 0;
}


// img = video_read_frame( n);
int video_read_frame( int slot, int pos);
int Frame2Mat(AVFrame *pFrame, int nPos, int nRow, int nCol, uint8_t * sci_buf);

int int_video_read_frame(char *fname)
{
	int l1, m1, n1;
	int l2, m2, n2;
	m2 = n2 = 1;
	int ret;
	
	CheckRhs(1,1) ;
	CheckLhs(1,1) ;
	GetRhsVar(1, "d", &m1, &n1, &l1); 
	ret = video_read_frame( (int)*stk(l1), 2);
	if (ret<0){
		sciprint(  "%s: Can not read video file.\r\n", fname);
		// return -1;
	}
	LhsVar(1) = 2;
	return 0;
}

int video_read_frame( int slot, int pos)
{
	// set up enviroment from table
	AVFormatContext *pFormatCtx = OpenedV[slot].pFormatCtx ;
  AVCodec	*pCodec = OpenedV[slot].pCodec;
	uint8_t *buffer = OpenedV[slot].buffer;
  AVFrame *pFrameRGB = OpenedV[slot].pFrameRGB;
  AVFrame *pFrame = OpenedV[slot].pFrame; 
	int videoStream = OpenedV[slot].videoStream;
	AVCodecContext  *pCodecCtx =  OpenedV[slot].pCodecCtx;
	uint8_t * sci_buf =  OpenedV[slot].sci_buf;
	// local vars
	int i = 0;
	AVPacket packet;
	int frameFinished;

	if (!pFormatCtx)
		return -1;
	// Read frames and save first five frames to disk
	frameFinished = 0;
	while ( frameFinished==0 ) {
		if (av_read_frame(pFormatCtx, &packet)>=0) { //get a frame
			// Is this a packet from the video stream?
			if(packet.stream_index==videoStream){
				// Decode video frame
				avcodec_decode_video(pCodecCtx, pFrame, &frameFinished, packet.data, packet.size);
				
				// Did we get a video frame?
				if(frameFinished){
					// Convert the image from its native format to RGB
					img_convert((AVPicture *)pFrameRGB, PIX_FMT_RGB24, 
					    (AVPicture*)pFrame, pCodecCtx->pix_fmt, pCodecCtx->width, 
					    pCodecCtx->height);
					// Frame2Mat
					Frame2Mat( pFrameRGB, pos,  pCodecCtx->height, pCodecCtx->width, sci_buf );
				}
			}
			// Free the packet that was allocated by av_read_frame
			av_free_packet(&packet);
	
		}else { 
			// end of file, return the last frame, bug? if no frame is readed
			// memset( sci_buf ,0, pCodecCtx->height * pCodecCtx->width * 3);
			Frame2Mat( pFrameRGB, pos,  pCodecCtx->height, pCodecCtx->width, sci_buf );
			frameFinished = 1;
		}
	} // end of while
	
	// dont forget to updata cur_frame
	OpenedV[slot].cur_frame++;
	return 0;
}

int Frame2Mat(AVFrame *pFrame, int nPos, int nRow, int nCol, uint8_t * sci_buf)
{
  static char *Str[]= { "hm","dims","entries"}; 
  int m1=1,n1=3;
  int mL=3,nL=1,lL, un=1;
  const int nCh = 3;
  int d[3];
  SciIntMat Dims;
  SciIntMat IntData;

  //create SciIntMat for dimentional information
  Dims.m = 1;
  Dims.n = 3;
  Dims.l = -1; /* dimensions matrix not in scilab stack */
  Dims.it = I_INT32;
	Dims.D = d;
	d[0] = nRow;
	d[1] = nCol;
	d[2] = nCh;

/*	
  Dims.D = malloc(Dims.n * sizeof(int)); //dim data
  if (!Dims.D)  {
    Scierror(999,"Unable to alloc memory for the image\n");
    return -1;
  }
  IC_INT32(Dims.D)[0] = nRow;
  IC_INT32(Dims.D)[1] = nCol;
  IC_INT32(Dims.D)[2] = nCh;
*/
	
  //create SciIntMat for the image data
  IntData.m = nRow ;
  IntData.n = nCol * nCh;
  IntData.l = -1; /* dimensions matrix not in scilab stack */
  IntData.it = I_UCHAR;//nType;
	IntData.D = sci_buf;
	
	// copy data	
  char * pSrc = (char*)pFrame->data[0];
  char * pDst = (char*)IntData.D;
	int ch, col, row;
	int widthStep = pFrame->linesize[0];

  for(ch = 0; ch < nCh ; ch++){
    for(col =0; col < nCol; col++){
      for(row = 0; row < nRow; row++){
				// (x,y) = (col, row)
				*pDst++ = * (pSrc + row*widthStep + col*nCh + ch);
			}
		}
	}
	
  CreateVar(nPos,"m", &mL, &nL, &lL);
  CreateListVarFromPtr(nPos,1,"S", &m1, &n1, Str);
  CreateListVarFromPtr(nPos,2,"I",&(Dims.m), &(Dims.n), &Dims); 
  CreateListVarFromPtr(nPos,3,"I",&(IntData.m), &(IntData.n), &IntData ); 

// free(Dims.D);
  return 0;
}

// video_close_all();
int int_video_close_all(char *fname)
{
	CheckRhs(0,0) ;
	CheckLhs(0,1) ;
	
	// for each opened video
	for (int n = 0; n < MAX_VIDEO_FILE_NUM; n++){
		if( (OpenedV[n].pFormatCtx) ){
		  video_close( n );
		}
  }
	return 0;
}

// [frame_rate, frame_num] = video_info( n )
int video_info( int slot, double *prate, double *pnum );

int int_video_info( char *fname)
{
	int l1, m1, n1;
	int l2, m2, n2;
	int l3, m3, n3;
	int ret;
	double frame_rate, frame_num;
	
	CheckRhs(1,1) ;
	CheckLhs(2,2) ;
	GetRhsVar(1, "d", &m1, &n1, &l1); 
	
	// return info
	m2 = 1; n2 = 1; m3=1; n3=1;
	CreateVar( 2, "d", &m2, &n2, &l2);
	CreateVar( 3, "d", &m3, &n3, &l3);
	int slot = (int) *stk(l1);
	*stk(l2) = OpenedV[slot].frame_rate;
	*stk(l3) = OpenedV[slot].frame_num;
	LhsVar(1) = 2;
	LhsVar(2) = 3;
	return 0;
}

/*  useless now
int video_info( int slot, double *prate, double *pnum )
{
	// set up enviroment from table
	AVFormatContext *pFormatCtx = OpenedV[slot].pFormatCtx ;
	int videoStream = OpenedV[slot].videoStream;
	
	if (pFormatCtx==0) 
		return -1;	// not open or close, return 
		
	// get information
	AVStream *st = 	pFormatCtx->streams[videoStream];
	if(st->r_frame_rate.den && st->r_frame_rate.num)
		*prate =  av_q2d(st->r_frame_rate);
	else
		*prate =  1/av_q2d(st->codec->time_base);
	
	*pnum = (*prate) * pFormatCtx->duration / AV_TIME_BASE;
	return 0;
}
*/


/*
added global vars:
	cur_frame, frame_num, frame_rate

skip_frame(), for frame_num & seek_to
get_num, skip all the frames of a video, get frame_num;
cur_frame, file pointor
*/



/*	
seek_to( int frame )
{
	fix the param;
 if (av_seek<=0){
 	if (cur_frame > frame ) // have to reopen
 	  close()
 	  reopen()
 	}
 	skip(frames)
 	if ( reopened)
 		save enviroment
 }
 return 0;
}
*/

int seek_to( int slot, double frame )
{
	// set up enviroment from table
	AVFormatContext *pFormatCtx = OpenedV[slot].pFormatCtx ;
	AVCodec	*pCodec = OpenedV[slot].pCodec;
	AVFrame *pFrame = OpenedV[slot].pFrame; 
	AVCodecContext  *pCodecCtx =  OpenedV[slot].pCodecCtx;
	int videoStream = OpenedV[slot].videoStream;
	long frame_num = OpenedV[slot].frame_num;
	long cur_frame = OpenedV[slot].cur_frame;
	double frame_rate = OpenedV[slot].frame_rate;

	if (!pFormatCtx) return -1;
	// fix the input
	if (frame<0)			frame = 0;
	if (frame>frame_num)		frame = frame_num;

	AVStream *st = 	pFormatCtx->streams[videoStream];
	// time = ( frame/frame_rate) / time_base;
	double time = frame/ frame_rate * st->time_base.den / st->time_base.num;
	
	sprintf( buf, "time =  %lf", time );
	debug( buf );
	
	//if (av_seek_frame(pFormatCtx, -1, time, AVSEEK_FLAG_ANY) < 0){
	if (1 ){
		// Seeking is not properly sorted with some codecs
		debug("Seeking is not properly sorted with the codecs");
		
		sprintf(buf, " if( %d> %d) ", (int)cur_frame, (int)frame );
		debug( buf);
		
		if ( cur_frame > frame ){
			debug( " have to reopen the file" );
			
			// close the file and starting again, stepping through the frames
			char filename[1024];
			strcpy( filename, pFormatCtx->filename );
			// close
			avcodec_close(pCodecCtx);
			av_close_input_file(pFormatCtx);
			
			// reopen
			// Open video file
			if(av_open_input_file(&pFormatCtx, filename, NULL, 0,  NULL)!=0)
				return -1; // Couldn't open file
		
			// Retrieve stream information
			if ( av_find_stream_info(pFormatCtx)<0 )
			    return -1; // Couldn't find stream information
			
			// Find the first video stream
			videoStream=-1;
			for(int i=0; i<pFormatCtx->nb_streams; i++)
			    if(pFormatCtx->streams[i]->codec->codec_type==CODEC_TYPE_VIDEO){
			        videoStream=i;
			        break;
			    }
			if(videoStream==-1)
				return -1; // Didn't find a video stream
		
			// Get a pointer to the codec context for the video stream
			pCodecCtx= pFormatCtx->streams[videoStream]->codec;
			
			// Find the decoder for the video stream
			pCodec=avcodec_find_decoder(pCodecCtx->codec_id);
			if(pCodec==NULL)
			    return -1; // Codec not found
			
			// Open codec
			if(avcodec_open(pCodecCtx, pCodec)<0)
			    return -1; // Could not open codec
		}// end of if for reopen
		
		debug ("skip some frames");
		// skip some frames
		for (int i = cur_frame; i< frame && skip_frame( pFormatCtx, pCodecCtx, pFrame, videoStream )>= 0 ; i++ )
		{	} // skipping
		
		// save enviroment now
		if ( cur_frame > frame ){
			debug("save enviroment");
			// save enviroment
			OpenedV[slot].pFormatCtx = pFormatCtx ;
			OpenedV[slot].pCodec = pCodec;
			OpenedV[slot].videoStream = videoStream;
			OpenedV[slot].pCodecCtx = pCodecCtx;
			
		}
	} // end of can not seek 
	
	OpenedV[slot].cur_frame = frame; // <==updata cur_frame

	debug(" return from seek to " );
	
	return 0;
}

// int seek_to( int slot, double frame )
// ret = video_seek( n, frame );
int int_seek_to( char *fname )
{
	int l1, m1, n1;
	int l2, m2, n2;
	int l3, m3, n3;
	int ret;
	
	CheckRhs(2,2) ;
	CheckLhs(1,1) ;
	GetRhsVar(1, "d", &m1, &n1, &l1); 
	GetRhsVar(2, "d", &m2, &n2, &l2); 

	// return info
	m3 = 1; n3 = 1;
	CreateVar( 3, "d", &m3, &n3, &l3);
	*stk(l3) = 0;	
	if ( seek_to( (int)*stk(l1), *stk(l2) ) < 0 ){
		sciprint(" error when seeking, the opened file may be useless!\n" );
		*stk(l3) = -1; // error
	}

	LhsVar(1) = 3;
	return 0;	
}


};// extern "C"

