//   This file is part of Scilab Aurora 
//	( http://scilab-aurora.sourceforge.net/ )
//   Copyright (C) 2006  XMU Scilab Aurora Team
//   Authors: (Scilab Aurora Team)
//     Jiang Fengbin <jonbo@126.com>
//     Lin Chuanwen <fjjllcw695@163.com>
//     Xiao Wenqing <pblz.xm@gmail.com>
//     Yang Kaigui  <justforward.xm@gmil.com>
//     Zheng Jiemin 
//     Zhou Xiyue   <zxy_0725@163.com>
//     Lin Shuyang  <macroxmu@gmail.com>
//   This program is licensed under GPL License;
//



// video write : wriet using ffmpeg
// by Jonbo
// n =video_open_w( filename, bit_rate, width, height,  frame_rate)
// video_write_frame( img, n)
// video_close_w(n)
// video_close_all_w()
extern "C" {
	
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "avformat.h"
#include "stack-c.h"

#define STREAM_PIX_FMT PIX_FMT_YUV420P /* default pix_fmt */

extern int already_registered;		// register only once

// opened video file table
const int MAX_VIDEO_FILE_NUM = 20;

typedef struct OpenedWri{
	AVFormatContext *oc;
	AVStream *video_st;

	AVFrame *picture;
	AVFrame *tmp_picture;
	uint8_t *video_outbuf;
	int  video_outbuf_size;
} OpenedWri;

OpenedWri OpenedW[MAX_VIDEO_FILE_NUM];

inline void debug(char *msg)
{
//	FILE *fp = fopen("debug.txt", "a");
//	fprintf( fp, "%s\n", msg );
//	fclose( fp );
}


// add a video output stream, return NULL if error
AVStream *add_video_stream(AVFormatContext *oc, int codec_id, 
		int bit_rate, int width, int height, int frame_rate )
{
    AVCodecContext *c;
    AVStream *st = 0;

    st = av_new_stream(oc, 0);
    if (!st) {
        sciprint("Could not alloc stream\n");
        return NULL;
    }

    c = st->codec;
    c->codec_id = (CodecID) codec_id;
    c->codec_type = CODEC_TYPE_VIDEO;
    c->bit_rate = bit_rate;  // 400000;
    /* resolution must be a multiple of two */
    if ( width&1 ) width++;
    if ( height&1 ) height++;
    c->width = width;
    c->height = height;
    c->time_base.den = frame_rate;   // STREAM_FRAME_RATE;
    c->time_base.num = 1;
    c->gop_size = 12; /* emit one intra frame every twelve frames at most */
    c->pix_fmt = STREAM_PIX_FMT;
    if (c->codec_id == CODEC_ID_MPEG2VIDEO) {
        /* just for testing, we also add B frames */
        c->max_b_frames = 2;
    }
    if (c->codec_id == CODEC_ID_MPEG1VIDEO){
        /* needed to avoid using macroblocks in which some coeffs overflow
           this doesnt happen with normal video, it just happens here as the
           motion of the chroma plane doesnt match the luma plane */
        c->mb_decision=2;
    }
    // some formats want stream headers to be seperate
    if(!strcmp(oc->oformat->name, "mp4") ||
    	 !strcmp(oc->oformat->name, "mov") || 
    	 !strcmp(oc->oformat->name, "3gp") )
        c->flags |= CODEC_FLAG_GLOBAL_HEADER;
    return st;
}

AVFrame *alloc_picture(int pix_fmt, int width, int height)
{
    AVFrame *picture;
    uint8_t *picture_buf;
    int size;

    picture = avcodec_alloc_frame();
    if (!picture)
        return NULL;
    size = avpicture_get_size(pix_fmt, width, height);
    picture_buf = (uint8_t *)malloc(size);
    if (!picture_buf) {
        av_free(picture);
        return NULL;
    }
    avpicture_fill((AVPicture *)picture, picture_buf, pix_fmt, width, height);
    return picture;
}

int video_open_w( int slot, char *filename, int bit_rate, int width, int height, int frame_rate)
{
	// global vars
	AVFormatContext *oc = 0;
	AVStream *video_st = 0;
	AVFrame *picture = 0;
	AVFrame *tmp_picture = 0;
	uint8_t *video_outbuf = 0;
	int video_outbuf_size = 0;
	// local vars
	AVOutputFormat *fmt = 0;
	AVCodec *codec = 0;
	AVCodecContext *c = 0;

	/* initialize libavcodec, and register all codecs and formats */
	if (already_registered==0){
		av_register_all();
		already_registered = 1;
	}

	// ******** part 1 ******** format context
	
	/* auto detect the output format from the name. default is  mpeg. */
	fmt = guess_format(NULL, filename, NULL);
	if (!fmt) {
		sciprint("Could not deduce output format from file extension: using MPEG.\n");
		fmt = guess_format("mpeg", NULL, NULL);
	}
	if (!fmt) {
		sciprint("Could not find suitable output format\n");
		goto _ERR;
	}

	/* allocate the output media context */
	oc = av_alloc_format_context();
	if (!oc) {
		sciprint( "Memory error\n");
		goto _ERR;
	}
	oc->oformat = fmt;
	snprintf(oc->filename, sizeof(oc->filename), "%s", filename);
	

	// ******** part 2 ******** video stream
	
	/* add the video stream using the default format codecs and initialize the codecs */
	if (fmt->video_codec == CODEC_ID_NONE) {
		sciprint("the format can not have video streams!\n");
		// <== bug? release oc?
		goto _ERR;
	}
	video_st = add_video_stream(oc, fmt->video_codec, bit_rate, width, height, frame_rate);
	if ( !video_st ){ 
		sciprint("error when add video stream!\n");
		goto _ERR;
	}
		
	/* set the output parameters (must be done even if no parameters). */
	if (av_set_parameters(oc, NULL) < 0) {
		sciprint( "Invalid output format parameters\n");
		goto _ERR;
	}
	dump_format(oc, 0, filename, 1);
	

	// ******** part 3 ******** video codec

	c = video_st->codec;
  /* find the video encoder */
	codec = avcodec_find_encoder(c->codec_id);
	if (!codec) {
		sciprint("codec not found\n");
		goto _ERR;
	}
	
	/* open the codec */
	if (avcodec_open(c, codec) < 0) {
		sciprint("could not open codec\n");
		goto _ERR;
	}


	// ******** part 4 ******** buffers
	
	// output buffer
	if (!(oc->oformat->flags & AVFMT_RAWPICTURE)) {
		video_outbuf_size = 200000;
		video_outbuf = (uint8_t*) malloc(video_outbuf_size);
		if (!video_outbuf){
			sciprint("Could not outbuf\n");
			goto _ERR;
		}
	}
	
	/* allocate the encoded raw picture */
	picture = alloc_picture(c->pix_fmt, c->width, c->height);
	if (!picture) {
		sciprint("Could not allocate picture\n");
		goto _ERR;
	}
	
	// the tmp_picture is used to store the RGB pic from scilab
	tmp_picture = alloc_picture(PIX_FMT_BGR24, c->width, c->height);	
	if (!tmp_picture) {
		sciprint("Could not allocate temporary picture\n");
		goto _ERR;
	}
	
	
	// ******** part 5 ******** output file
	
	/* open the output file, if needed */
	if ((fmt->flags & AVFMT_NOFILE)) {
		sciprint( "not supported format!\n");
		goto _ERR;		
	}
	if (url_fopen(&oc->pb, filename, URL_WRONLY) < 0) {
		sciprint("Could not open '%s'\n", filename);
		goto _ERR;
	}

	// write the stream header, if any 
	av_write_header(oc);

	// ******** part 6 ******** save enviroment
	OpenedW[slot].oc = oc;
	OpenedW[slot].video_st = video_st;
	OpenedW[slot].picture = picture;
	OpenedW[slot].tmp_picture = tmp_picture;
	OpenedW[slot].video_outbuf = video_outbuf;
	OpenedW[slot].video_outbuf_size = video_outbuf_size;
	return 0;
	
	// *** part 7 ******** error handle
_ERR:
/*
	// release alloc objects
	if (picture && picture->data[0])
		av_free(picture->data[0]);
	if (picture)
		av_free(picture);
	if ( tmp_picture && tmp_picture->data[0] )
		av_free(tmp_picture->data[0]);
	if (tmp_picture)
		av_free(tmp_picture);
	if (video_outbuf)
		av_free(video_outbuf);
	if ( video_st && video_st->codec )
		avcodec_close(video_st->codec);
	// free the streams
	if (oc){
//		for(int i = 0; i < oc->nb_streams; i++) {
//			av_freep(&oc->streams[i]->codec);
//			av_freep(&oc->streams[i]);
//		}
		av_free(oc);
	}
*/

	OpenedW[slot].oc = 0;
	OpenedW[slot].video_st = 0;
	OpenedW[slot].picture = 0;
	OpenedW[slot].tmp_picture = 0;
	OpenedW[slot].video_outbuf = 0;
	OpenedW[slot].video_outbuf_size = 0;
	return -1;
}


int video_close_w( int slot )
{
	// load enviroment
	AVFormatContext *oc = OpenedW[slot].oc;
	AVStream *video_st = OpenedW[slot].video_st;
	AVFrame *picture = OpenedW[slot].picture;
	AVFrame *tmp_picture = OpenedW[slot].tmp_picture;
	uint8_t *video_outbuf = OpenedW[slot].video_outbuf;
	int  video_outbuf_size = OpenedW[slot].video_outbuf_size;
	
	if (!oc)
		return -1;
	// release alloc objects
	avcodec_close(video_st->codec);
	av_free(picture->data[0]);
	av_free(picture);
	av_free(tmp_picture->data[0]);
	av_free(tmp_picture);
	av_free(video_outbuf);
	
	/* write the trailer, if any */
	av_write_trailer(oc);
	
	/* free the streams */
	for(int i = 0; i < oc->nb_streams; i++) {
		av_freep(&oc->streams[i]->codec);
		av_freep(&oc->streams[i]);
	}
	
	/* close the output file */
	if (!(oc->oformat->flags & AVFMT_NOFILE)) {
		url_fclose(&oc->pb);
	}
	
	/* free the stream */
	av_free(oc);
	return 0;
}


// check types & get width, height ,Data
int get_data( int pos, int *pWidth, int *pHeight, char **ppData)
{
  char ** pStr;
  int * pListHeader;
  int * pDataHeader;
  int m1, n1, m2, n2, m3, n3;
  SciIntMat Dims;
  SciIntMat IntMat;
  int nWidth, nHeight, nCh;

	debug(" checking is it a hypermatrix" );
  //check whether the the argument is a hypermatrix
	if( VarType(pos) != 17) // not a hypermatrix
  	return -1;
	GetListRhsVar (pos, 1, "S", &m1, &n1, &pStr);
  if( m1 !=1 || n1 !=3)
  	return -1;
  if( strcmp(pStr[0], "hm") != 0 || strcmp(pStr[1], "dims") != 0 || strcmp(pStr[2], "entries") != 0)
  	return -1;

	debug("get the dimension information");
  //get the dimension information, it's stored in the second element of the mlist
  GetListRhsVar(pos, 2, "I", &m2, &n2, &Dims);
  if( m2*n2 !=3)
  	return -1;
  nHeight = IC_INT32(Dims.D)[0];
  nWidth  = IC_INT32(Dims.D)[1];
  nCh = IC_INT32(Dims.D)[2];
  if (nCh!=3)
  	return -1;
  
  debug("get mlist data structure");	
  //get mlist data structure
  pListHeader = (int*)GetData(pos);
  //data is stored in the third element
 //data struct is get first
  pDataHeader = (int*)( ((char*)pListHeader) + 24 + (pListHeader[4]-1)*8 );
  //the data type is store in the first 4 bytes as a integer
  if ( pDataHeader[0]!=8 )	// data type must be integer
  	return -1;
  
  debug(" get dimesion");	
	GetListRhsVar(pos, 3, "I", &m3, &n3, &IntMat);
	m3 = IntMat.m; 	n3 = IntMat.n;
	if (IntMat.it!=I_UCHAR)
  	return -1;
	if(m3*n3 != nWidth * nHeight * nCh) 	//check dimension 
  	return -1;
	// pData = IntMat.D;

	// now we got nWidth, nHeight, pData
	*pWidth = nWidth;
	*pHeight = nHeight;
	*ppData = (char*)IntMat.D;
 
EXIT:
  debug(" return from get_data" );
//  FreeRhsSVar(pStr);
  // <==bug here?, do I have to free other vars?
  return 0;
}


int get_image_from_scilab( AVFrame *pic, int pos, int width, int height )
{
	int src_width, src_height;
	char *pData;
	
	// get data & params from scilab
	if ( get_data(pos, &src_width, &src_height, &pData)<0 ){
		sciprint("Not a RGB image?");
		return -1;	
	}
	if ( src_width!=width || src_height!=height ){
		sciprint(" something wrong with input frame size!\n");
		return -1;
	}
	
	// convert to AVFrame
	uint8_t *pDst = pic->data[0];
	uint8_t *pSrc = (uint8_t*)pData;
	int row, col, ch;
	int widthStep = pic->linesize[0];
	for(ch = 0; ch < 3 ; ch++)
		for(col =0; col < width; col++)
			for(row = 0; row < height; row++){
				// (x,y)=(col,row)
				*(pDst + widthStep*row + col*3 + ch) = *pSrc++;
			}
	
	return 0;
}

int video_write_frame( int slot, int pos)
{
	// load enviroment
	AVFormatContext *oc = OpenedW[slot].oc;
	AVStream *st = OpenedW[slot].video_st;
	AVFrame *picture = OpenedW[slot].picture;
	AVFrame *tmp_picture = OpenedW[slot].tmp_picture;
	uint8_t *video_outbuf = OpenedW[slot].video_outbuf;
	int  video_outbuf_size = OpenedW[slot].video_outbuf_size;
	// local vars
	int out_size, ret;
	AVCodecContext *c;
  c = st->codec;
  
	// get image from scilab & convert
	if (c->pix_fmt != PIX_FMT_RGB24){
	   // we must convert it to the codec pixel format if needed
	   get_image_from_scilab( tmp_picture, pos, c->width, c->height );
	   img_convert((AVPicture *)picture, c->pix_fmt,
	               (AVPicture *)tmp_picture, PIX_FMT_RGB24,
	               c->width, c->height);
	} else {
	   // fill_yuv_image(picture, frame_count, c->width, c->height);
   	 get_image_from_scilab( tmp_picture, pos, c->width, c->height );
	}
	
	// write frame
	if (oc->oformat->flags & AVFMT_RAWPICTURE) {
		// raw video case. 
		AVPacket pkt;
		av_init_packet(&pkt);
		pkt.flags |= PKT_FLAG_KEY;
		pkt.stream_index= st->index;
		pkt.data= (uint8_t *)picture;
		pkt.size= sizeof(AVPicture);
		ret = av_write_frame(oc, &pkt);

	} else {
		// encode the image
		out_size = avcodec_encode_video(c, video_outbuf, video_outbuf_size, picture);
		// if zero size, it means the image was buffered
		if (out_size > 0) {
			AVPacket pkt;
			av_init_packet(&pkt);
			pkt.pts= av_rescale_q(c->coded_frame->pts, c->time_base, st->time_base);
			if(c->coded_frame->key_frame)
				pkt.flags |= PKT_FLAG_KEY;
			pkt.stream_index= st->index;
			pkt.data= video_outbuf;
			pkt.size= out_size;
		
			/* write the compressed frame in the media file */
			ret = av_write_frame(oc, &pkt);
		} else {
			ret = 0;
		}
	}

	if (ret != 0) {
		sciprint("Error while writing video frame\n");
		return -1;
	}
	return 0;
}

// n =video_open_w( filename, bit_rate, width, height,  frame_rate)
int int_video_open_w( char *fname )
{
	int m1, n1, l1;
	int m2, n2, l2;
	int m3, n3, l3;
	int m4, n4, l4;
	int m5, n5, l5;
	int m6, n6;
	int n;
	int *pret = &n;

	CheckRhs(5,5) ;
	CheckLhs(1,1) ;
	GetRhsVar(1, "c", &m1, &n1, &l1); 
	GetRhsVar(2, "d", &m2, &n2, &l2); 
	GetRhsVar(3, "d", &m3, &n3, &l3); 
	GetRhsVar(4, "d", &m4, &n4, &l4); 
	GetRhsVar(5, "d", &m5, &n5, &l5); 

	// find a empty slot
	for ( n = 0; n < MAX_VIDEO_FILE_NUM; n++){
		if( !(OpenedW[n].oc))
			break;
  }
	if( n ==  MAX_VIDEO_FILE_NUM){
		sciprint("%s: Too many video files  opened.  close some files .\r\n", fname);
		// return -1;
		n = -1; goto _ERR;
	}
	
	if ( video_open_w( n, cstk(l1), (int)*stk(l2), (int)*stk(l3), (int)*stk(l4), (int)*stk(l5)) < 0 ){
		sciprint(  "%s: Can not open video file.\r\n", fname);
		// return -1;
		n = -1;
	}

_ERR:
	// return handle
	m6=n6=1;
	CreateVarFromPtr(6, "i", &m6, &n6, &pret);
	LhsVar(1) = 6;
	return 0;
}


// video_close_w(n)
int int_video_close_w( char *fname )
{
	int l1, m1, n1;
	CheckRhs(1,1) ;
	CheckLhs(0,1) ;
	GetRhsVar(1, "d", &m1, &n1, &l1); 
	
	int slot = (int)*stk(l1);
	if ( slot<0 || slot> MAX_VIDEO_FILE_NUM || video_close_w(slot ) <0 ){
		sciprint("error when closing, maybe the file is not open?");
		// return -1;
	}
	return 0;
}

// video_close_all()
int int_video_close_all_w( char *fname)
{
	CheckRhs(0,0) ;
	CheckLhs(0,1) ;

	// for each opened video
	for (int n = 0; n < MAX_VIDEO_FILE_NUM; n++){
		if( (OpenedW[n].oc) ){
		  video_close_w( n );
		}
  }
	return 0;
}

// video_write_frame(img, n);
int int_video_write_frame( char *fname )
{
	int m2, n2, l2;
	int n;
	CheckRhs(2,2);
	CheckLhs(1,1);
	
	GetRhsVar(2, "d", &m2, &n2, &l2); 
	n = (int)*stk(l2);

	int m3, n3, l3;
	m3 =n3 = 1;
	CreateVar( 3, "d", &m3, &n3, &l3);
	*stk(l3) = 0;	
	
	if ( n<0 || n>MAX_VIDEO_FILE_NUM || !OpenedW[n].oc ){
		sciprint("the file is not opened\n");
		// return -1;
		*stk(l3) = -1;	goto _ERR;
	}
	
	if ( video_write_frame( n, 1) < 0 ){
		sciprint(" error when write frame" );
		*stk(l3) = -1;	
		// return -1;
	}

_ERR:
	LhsVar(1) = 3;
	return 0;
}
	

};

